@@ -8,7 +8,10 @@
 # copied, modified, or distributed except according to those terms.
 
 import logging
+import os
+import platform
 import sys
+import time
 from getpass import getpass
 from numbers import Number
 from threading import Event, Timer
@@ -20,6 +23,8 @@ from pynitrokey.confconsts import (
     GH_ISSUES_URL,
     LOG_FN,
     SUPPORT_EMAIL,
+    SUPPORT_URL,
+    UDEV_URL,
     VERBOSE,
     Verbosity,
 )
@@ -75,6 +80,15 @@ class ProgressBar:
             self.bar.close()
 
 
+class DownloadProgressBar(ProgressBar):
+    """
+    Helper class for progress bars for downloading a file.
+    """
+
+    def __init__(self, desc: str) -> None:
+        super().__init__(desc=f"Download {desc}", unit="B", unit_scale=True)
+
+
 class Timeout(object):
     """
     Utility class for adding a timeout to an event.
@@ -102,6 +116,41 @@ class Timeout(object):
             self.timer.cancel()
 
 
+class Try:
+    """Utility class for an execution of a repeated action with Retries."""
+
+    def __init__(self, i: int, retries: int) -> None:
+        self.i = i
+        self.retries = retries
+
+    def __str__(self) -> str:
+        return f"try {self.i + 1} of {self.retries}"
+
+    def __repr__(self) -> str:
+        return f"Try(i={self.i}, retries={self.retries})"
+
+
+class Retries:
+    """Utility class for repeating an action multiple times until it succeeds."""
+
+    def __init__(self, retries: int, timeout: float = 0.5) -> None:
+        self.retries = retries
+        self.i = 0
+        self.timeout = timeout
+
+    def __iter__(self) -> "Retries":
+        return self
+
+    def __next__(self) -> Try:
+        if self.i >= self.retries:
+            raise StopIteration
+        if self.i > 0:
+            time.sleep(self.timeout)
+        t = Try(self.i, self.retries)
+        self.i += 1
+        return t
+
+
 # @todo: introduce granularization: dbg, info, err (warn?)
 #        + machine-readable
 #        + logfile-only (partly solved)
@@ -121,7 +170,7 @@ def local_print(*messages, **kwargs):
             logger.exception(item)
             passed_exc = item
             item = repr(item)
-            item = "Exception encountered: " + item
+            item = "\tException encountered: " + item
 
         # just a newline, don't log to file...
         elif item is None or item == "":
@@ -151,20 +200,30 @@ def local_critical(*messages, support_hint=True, ret_code=1, **kwargs):
         # @fixme: not the best solution
         STDOUT_PRINT = False
         local_print("listing all connected devices:")
-        from pynitrokey.cli import nitropy
 
-        nitropy.commands["ls"].callback()
+        try:
+            from pynitrokey.cli import nitropy
+
+            nitropy.commands["list"].callback()
+        except Exception:
+            local_print("Unable to list devices. See log for the details.")
+            logger = logging.getLogger()
+            logger.exception("Unable to list devices")
+
         STDOUT_PRINT = True
 
+        linux = "linux" in platform.platform().lower()
         local_print(
             "",
             "-" * 80,
             "Critical error occurred, exiting now",
-            "Unexpected? Is this a bug? Do you would like to get support/help?",
-            f"- You can report issues at: {GH_ISSUES_URL}",
-            f"- Writing an e-mail to: {SUPPORT_EMAIL} is also possible",
+            "Unexpected? Is this a bug? Would you like to get support/help?",
+            f"- You can report issues at: {SUPPORT_URL}",
+            f"- Writing an e-mail to {SUPPORT_EMAIL} is also possible",
             f"- Please attach the log: '{LOG_FN}' with any support/help request!",
-            "-" * 80,
+            f"- Please check if you have udev rules installed: {UDEV_URL}"
+            if linux
+            else "" "-" * 80,
             "",
         )
     sys.exit(ret_code)
@@ -192,6 +251,7 @@ class AskUser:
         repeat: int = 3,
         adapt_question=True,
         hide_input=False,
+        envvar: str = None,
     ):
 
         self.data = None
@@ -216,6 +276,7 @@ class AskUser:
         self.strict = strict
         self.repeat = repeat or 1
         self.hide_input = hide_input
+        self.envvar = envvar
 
     @classmethod
     def yes_no(cls, what: str, strict: bool = False):
@@ -244,7 +305,14 @@ class AskUser:
         )
 
     def ask(self):
-        answer = self.get_input()
+        answer = None
+        if self.envvar is not None:
+            fromvar = os.environ.get(self.envvar)
+            if fromvar is not None:
+                answer = fromvar
+
+        if answer is None:
+            answer = self.get_input()
 
         # handle plain input request first
         if not self.options:
