@@ -10,6 +10,7 @@
 from typing import List, Optional
 
 from .base import Nitrokey3Base
+from .bootloader import Nitrokey3Bootloader
 from .device import Nitrokey3Device
 
 VID_NITROKEY = 0x20A0
@@ -18,8 +19,19 @@ PID_NITROKEY3_BOOTLOADER = 0x42DD
 
 
 def list() -> List[Nitrokey3Base]:
-    return [device for device in Nitrokey3Device.list()]
+    devices: List[Nitrokey3Base] = []
+    devices.extend(Nitrokey3Bootloader.list())
+    devices.extend(Nitrokey3Device.list())
+    return devices
 
 
 def open(path: str) -> Optional[Nitrokey3Base]:
-    return Nitrokey3Device.open(path)
+    device = Nitrokey3Device.open(path)
+    bootloader = Nitrokey3Bootloader.open(path)
+    if device and bootloader:
+        raise Exception(f"Found multiple devices at path {path}")
+    if device:
+        return device
+    if bootloader:
+        return bootloader
+    return None
